package checkers.nonnull;

import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.*;

import checkers.quals.*;
import checkers.types.*;
import checkers.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;
import com.sun.tools.javac.tree.*;

/**
 * Adds support for the {@code @NonNull} type annotation to
 * {@link AnnotatedTypeFactory}. This means that the
 * {@link NonnullAnnotatedTypeFactory#getClass} and
 * {@link NonnullAnnotatedTypeFactory#getMethod} methods will regard some
 * inputs as {@code @NonNull} even if they do not carry a {@code @NonNull}
 * annotation. These include:
 *
 * <ul>
 *  <li>primitive, class, and enum types
 *  <li>{@code super} and {@code this}
 *  <li>the results of binary operations
 *  <li>literals (except the null literal) and arrays of literals
 * </ul>
 */
@Default("checkers.quals.NonNull")
public class NonnullAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Used to perform flow-sensitive nonnull analysis. */
    protected FlowVisitor flow;
    
    protected Elements elements;
    
    protected SourcePositions srcPos;

    /**
     * Creates an annotated type factory that treats certain inputs as having
     * {@code @NonNull} annotations even if they do not explicitly carry a
     * {@code @NonNull} annotation.
     *
     * @param env the {@link ProcessingEnvironment} to use for tree and type
     *        utilities
     * @param root the root of the syntax tree (used for TreePaths)
     * @param useFlow whether or not flow-sensitive nonnull analysis should be used
     * @param ignoreSameTypeCast whether or not redundant cast annotations are used
     * 
     * @see FlowVisitor
     */
    public NonnullAnnotatedTypeFactory(ProcessingEnvironment env, CompilationUnitTree root, boolean useFlow, boolean ignoreSameTypeCast) {
        super(env, root);

        @Nullable Elements elements = env.getElementUtils();
        assert elements != null; /*nninvariant*/
        this.elements = elements;

        super.ignoreSameTypeCast = ignoreSameTypeCast;

        @Nullable SourcePositions srcPos = trees.getSourcePositions();
        assert srcPos != null; /*nninvariant*/
        this.srcPos = srcPos;

        this.flow = new FlowVisitor(root, trees.getSourcePositions(), this);
        if (useFlow) {
            // Run a flow-sensitive analysis for "if (x != null)"-type expressions.
            flow.scan(root, null);
        }
    }

    @Override
    public AnnotatedClassType getClass(Tree tree) {

        // If the given tree is a method invocation, return the annotated type
        // of the method's return value.
        if (tree instanceof MethodInvocationTree) {
            AnnotatedClassType ret = getMethod(tree).getAnnotatedReturnType();
            ExecutableElement ex = (ExecutableElement)ret.getElement();
            if (ex.getReturnType().getKind().isPrimitive())
                ret.include(NonNull.class);

            long pos = srcPos.getStartPosition(root, tree);
            for (FlowScope fs : flow.getResults()) 
                if (fs.contains(ex, pos)) 
                    ret.annotate(annotations.createAnnotation(NonNull.class.getName(),
                                AnnotationLocation.RAW),
                            annotations.createAnnotation(Nullable.class.getName(),
                                AnnotationLocation.RAW));

            applyNullable(tree, ret);

            return ret;
        }

        // Get the original annotated type, to which @NonNull annotations
        // may be added by the remainder of this method.
        AnnotatedClassType type = super.getClass(tree);

        @Nullable Tree t = TreeInfo.skipParens((JCTree)tree);
        assert t != null; /*nninvariant*/

        // Check results of flow-sensitive analysis.
        @Nullable Element tElt = InternalUtils.symbol(tree);

        long pos = srcPos.getStartPosition(root, tree);
        for (FlowScope fs : flow.getResults())
            if (fs.contains(tElt, pos)) 
                type.annotate(annotations.createAnnotation(NonNull.class.getName(),
                            AnnotationLocation.RAW),
                        annotations.createAnnotation(Nullable.class.getName(),
                            AnnotationLocation.RAW));


        if (tree instanceof TypeCastTree) {

            TypeCastTree tc = (TypeCastTree)tree;
            if (tc.getType().getKind() == Tree.Kind.PRIMITIVE_TYPE)
                type.include(NonNull.class);

        } else if (tree instanceof IdentifierTree) {

            IdentifierTree id = (IdentifierTree)tree;
            if (id.getName().contentEquals("super") || id.getName().contentEquals("this"))
                type.include(NonNull.class);

            @Nullable Element elt = InternalUtils.symbol(id);
            
            if (elt != null && /*nnbug*/
                    (elt.getKind() == ElementKind.CLASS || elt.getKind() == ElementKind.ENUM))
                type.include(NonNull.class);

        } else if (tree instanceof BinaryTree) {
            type.include(NonNull.class);
        } else if (tree instanceof MemberSelectTree) {

            MemberSelectTree ms = (MemberSelectTree)tree;
            @Nullable Element elt = InternalUtils.symbol(ms.getExpression());
            @Nullable Element field = InternalUtils.symbol(ms);
            if (field != null && "class".equals(field.getSimpleName().toString())) { /*nnbug*/
                TypesUtils tUtils = new TypesUtils(env);
                for (AnnotationLocation l : tUtils.allLocations(field.asType()))
                    type.include(annotations.createAnnotation(NonNull.class.getName(), l));
            }
        }

        Element typeElt = type.getElement();
        if (typeElt instanceof VariableElement 
                && typeElt.asType().getKind().isPrimitive())
            type.include(NonNull.class);

        switch (tree.getKind()) {
        case NEW_ARRAY:
            NewArrayTree nat = (NewArrayTree)tree;
            if (nat.getInitializers() != null) {
            boolean allNN = true;
            for (ExpressionTree et : nat.getInitializers()) {
                if (!getClass(et).hasAnnotationAt(NonNull.class, AnnotationLocation.RAW)) {
                    allNN = false;
                    break;
                }
            }
            if (allNN) {
                AnnotationLocation loc = AnnotationLocation.fromArray(new int[] { 0 });
                AnnotationData syn = annotations.createAnnotation(NonNull.class.getName(), loc);
                type.include(syn);
            }
            }
            // Don't break.
        case NEW_CLASS:
        case BOOLEAN_LITERAL:
        case DOUBLE_LITERAL:
        case FLOAT_LITERAL:
        case INT_LITERAL:
        case LONG_LITERAL:
        case CHAR_LITERAL:
        case STRING_LITERAL:
        case CONDITIONAL_AND:
        case CONDITIONAL_OR:
            type.include(NonNull.class);
            break;
        case NULL_LITERAL:
            type.annotate(annotations.createAnnotation(Nullable.class.getName(), AnnotationLocation.RAW));
            type.exclude(NonNull.class);
            break;
        case UNARY_MINUS:
            UnaryTree unary = (UnaryTree)tree;
            if (unary.getExpression() instanceof LiteralTree)
                type.include(NonNull.class);
        }

        applyNullable(tree, type);

        return type;
    }

    private void applyNullable(Tree tree, AnnotatedClassType type) {
        String aName = NonNull.class.getName();
        if (type.hasAnnotationAt(Nullable.class, AnnotationLocation.RAW))
            type.exclude(annotations.createAnnotation(aName, AnnotationLocation.RAW));
        TypesUtils tUtils = new TypesUtils(this.env);
        @Nullable TreePath path = this.trees.getPath(root, tree);
        if (path != null) {
            @Nullable TypeMirror theType = this.trees.getTypeMirror(path);
            if (theType != null) {
                for (AnnotationLocation loc : tUtils.allLocations(theType)) {
                    if (!type.hasAnnotationAt(Nullable.class, loc))
                        continue;
                    type.exclude(annotations.createAnnotation(aName, loc));
                }
            }
        }
    }

    @Override
    protected AnnotatedClassType receiverType(ExecutableElement method) {
        AnnotatedClassType type = super.receiverType(method);
        if (!method.getModifiers().contains(Modifier.STATIC))
            type.include(NonNull.class);
        return type;
    }

    @Override
    public AnnotatedMethodType getMethod(Tree tree) {
        AnnotatedMethodType type = super.getMethod(tree);

        // TODO *** remove this eventually and replace with a @SuppressWarnings mechanism

        // Make static final fields implicitly @Nonnull (so we don't get type
        // errors from things like System.out)
        if (tree instanceof MethodInvocationTree) {
            MethodInvocationTree mi = (MethodInvocationTree)tree;
            if (mi.getMethodSelect() instanceof MemberSelectTree) {
                MemberSelectTree ms = (MemberSelectTree)mi.getMethodSelect();
                @Nullable Element elt = InternalUtils.symbol(ms.getExpression());
                if (elt != null /*&& elt.getModifiers().contains(Modifier.FINAL)*/
                        && elt.getModifiers().contains(Modifier.STATIC))
                    type.getAnnotatedReceiverType().exclude(NonNull.class);
//                if (super.isOuterClassReference(ms))
//                    type.getAnnotatedReturnType().include(NonNull.class);
            }
        }

        ExecutableElement method = methodSymbol(tree);
        AnnotatedClassType select = methodSelect(tree);
        if (method != null && select != null && select.getElement() != null)
            updateMethodGenerics(method, tree, select, type);

        return type;
    }
}
